import numpy as np
import os
from mpi4py import MPI
from typing import Any, List
from ..parsers.PAFIParser import PAFIParser
from .LAMMPSWorker import LAMMPSWorker
from ..results.ResultsHolder import ResultsHolder

class PAFIWorker(LAMMPSWorker):
    """
        The hyperplane-constrained sampling run.
        results: ResultsHolder instance
            essentially a dictionary, defining both 
            input and output data- temeperature, reaction coordinate
            and other are defined, and others are filled.
            *Crucially, if a parameter is specified in results, 
            it overrides any value in the parameter file*
            Must have at least `ReactionCoordinate` and `Temperature` defined
        
        The PAFI workflow can be summarised as:
            1) Execute `PreRun` script
            2) Apply `fix_pafi` constraint at defined `ReactionCoordinate`
            3) Execute `PreTherm` script
            4) Thermalization for `ThermSteps` steps at `Temperature`
            5) Execute `constrained_average()` function
                In standard PAFI this runs for `SampleSteps` steps and 
                time averages the output of `fix_pafi`, as shown below.
                See https://docs.lammps.org/fix_pafi.html for details.
            6) Extract the average displacment from path if `PostDump==1`
            7) Minimize in-plane to check system returns to path.
                The check is the max per-atom displacement : `MaxJump`
                If `MaxJump` is larger than `MaxJumpMaxJumpThresh` then 
                the sample is retained but marked as `Valid=False`
            8) Execute `PostRun` script
        
        The `contrained_average()` function is therefore suitable for
        any form of hyperplane-constrained averaging.
        """
    
    def __init__(self, comm: MPI.Intracomm, 
                 parameters: PAFIParser, tag: int,
                 rank: int, roots: List[int]) -> None:
        super().__init__(comm, parameters, tag, rank, roots)
        
    
    def constrained_average(self,results:ResultsHolder)->ResultsHolder:
        """
        
        Parameters
        ----------
        results: ResultsHolder instance
            used to contain results and custom input paramaters
            we want these inputs to override self.parameters()
        
        Returns
        ----------
        ResultsHolder instance
            Returns the input data and all output data appended 
            as dictionary key,value pairs
        """
        # ensure results inputs should override self.parameters()
        parameters = lambda k: results(k)\
            if results.has_key(k) else self.parameters(k)
        fixname = self.setup_pafi_average(parameters("SampleSteps"),"avepafi")
        self.run_commands("run %d" % parameters("SampleSteps"))
        results = self.extract_pafi_data(results,fixname)
        
        return results
    
    def sample(self,results:ResultsHolder)->ResultsHolder:
        """
        Main sampling run.
        
        1) Execute `PreRun` script
        2) Apply `fix_pafi` constraint at defined `ReactionCoordinate`
        3) Execute `PreTherm` script
        4) Thermalization for `ThermSteps` steps at `Temperature`
        
        5) Execute some averaging function
            In standard PAFI this runs for `SampleSteps` steps and 
            time averages the output of `fix_pafi`, as shown below.
            See https://docs.lammps.org/fix_pafi.html for details.
        
        6) Extract the average displacment from path if `PostDump==1`
        7) Minimize in-plane to check system returns to path.
            The check is the max per-atom displacement : `MaxJump`
            If `MaxJump` is larger than `MaxJumpMaxJumpThresh` then 
            the sample is retained but marked as `Valid=False`
        8) Execute `PostRun` script

        
        Parameters
        ----------
        results: ResultsHolder instance
            used to contain results and custom input paramaters
            we want these inputs to override self.parameters()
        
        Returns
        ----------
        ResultsHolder instance
            Returns the input data and all output data appended 
            as dictionary key,value pairs
        """
        
    
        results = self.standard_pafi_pre_average(results)
        results = self.constrained_average(results)
        results = self.standard_pafi_post_average(results)

        return results
    

    def setup_pafi_average(self,ave_steps:int,fixname="avepafi")->str:
        """Helper function to establish PAFI average
        Parameters
        ----------
        
        ave_steps : int
            steps for ave/time
        fixname:str, optional
            the fix name, default "avepafi"
        Returns
        ---------
        str:
            the fix name
        """
        self.run_commands(f"""
            fix {fixname} all ave/time 1 {ave_steps} {ave_steps} f_pafi[*]
        """)
        return fixname
    
    def extract_pafi_data(self,results:ResultsHolder,
                          name:str="avepafi")->ResultsHolder:
        """Helper functions to extract PAFI data

        Parameters
        ----------
        results : ResultsHolder instance
            add data and returns
        name : str, optional
            fix name, default "avepafi"

        Returns
        -------
        ResultsHolder instance
        """
        res = {}
        fix_data = self.extract_fix(name,size=4)
        res['FreeEnergyGradient'] = -fix_data[0] * self.norm_t
        res['FreeEnergyGradientVariance'] = fix_data[1]**2 * self.norm_t**2 - res['FreeEnergyGradient']**2
        res['avePsi'] = fix_data[2]
        res['dXTangent'] = fix_data[3]
        results.set_dict(res)
        self.run_commands(f"unfix {name}")
        return results
    
    
    def standard_pafi_pre_average(self,results:ResultsHolder)->ResultsHolder:
        """Helper functions for standard PAFI

        Parameters
        ----------
        results : ResultsHolder instance
            custom input data overrides parameters
        
        Returns
        ----------
        results : ResultsHolder instance
            add data and returns custom inputs as well    
        """
        assert results.has_key("ReactionCoordinate")
        assert results.has_key("Temperature")
        
        self.kB = 8.617e-5 # where should I put this...

        # we want any input paramaters in results() to override parameters()
        parameters = lambda k: results(k) \
            if results.has_key(k) else self.parameters(k)
        
        # initialize on hyperplane
        r = results("ReactionCoordinate")
        T = results("Temperature")
        
        self.initialize_hyperplane(r,0.)
        # TODO: check order in public PAFI
        self.run_script("PreRun",results)
        self.initialize_hyperplane(r,T)
        n_atoms = self.get_natoms()

        # the PAFI fix
        gamma = parameters("Friction")
        overdamped = parameters("OverDamped")
        seed = self.parameters.randint()
        cmd = f"fix pafi all pafi __pafipath {T} {gamma} {seed} "
        cmd += f"overdamped {overdamped} com 1\nrun 0"
        self.run_commands(cmd)

        # pre minimize (optional)
        if parameters("PreMin"):
            min_steps = parameters("MinSteps")
            self.run_commands(f"""
                min_style fire
                minimize 0 0.0001 {min_steps} {min_steps}
            """)
        if parameters("PostMin"):
            self.change_x = -self.gather("x",1,3)
        
        # PreThermalize (optional)
        self.run_script("PreTherm",results)
        results.set("MinEnergy",self.get_energy())
        
        # establish temperature time average and thermalize
        pre_therm_hp = self.extract_fix("pafi",size=4)
        steps = parameters("ThermSteps")
        ave_steps = parameters("ThermWindow")
        f_T = "c_thermo_pe" if overdamped==1 else "c_thermo_temp"
        self.run_commands(f"""
            reset_timestep 0
            fix __ae all ave/time 1 {ave_steps} {steps} {f_T}
            run {steps}
        """)
        sampleT = self.extract_fix("__ae")
        
        if overdamped==1:
            sampleT = (sampleT-results("MinEnergy"))/1.5/n_atoms/self.kB
        results.set("preTemperature",sampleT)
        self.run_commands(f"""
            unfix __ae
            run 0""")
        
        # main sampling run
        steps = parameters("SampleSteps")
        self.run_commands(f"""
            reset_timestep 0
            fix __ae all ave/time 1 {steps} {steps} {f_T}
            """)
        if parameters("PostDump"):
            self.run_commands(f"""
            fix pafiax all ave/atom 1 {steps} {steps} x y z
            """)
        return results
    
    def standard_pafi_post_average(self,results:ResultsHolder)->ResultsHolder:
        """Helper functions for standard PAFI

        Parameters
        ----------
        results : ResultsHolder instance
            custom input data overrides parameters
        
        Returns
        ----------
        results : ResultsHolder instance
            add data and returns custom inputs as well    
        """
        # we want any input paramaters in results() to override self.parameters()
        parameters = lambda k: results(k) if results.has_key(k) else self.parameters(k)
        n_atoms = self.get_natoms()
        
        # get final temperature
        sampleT = self.extract_fix("__ae")
        if parameters("OverDamped")==1:
            sampleT = (sampleT-results("MinEnergy"))/1.5/n_atoms/self.kB
        results.set("postTemperature",sampleT)
        self.run_commands(f"""
            unfix __ae
            run 0""")
        
        # average positions
        if parameters("PostDump"):
            dx = self.gather("f_pafiax",type=1,count=3)
            dx[:,0] -= self.gather("f_ux",type=1,count=1).flatten()
            dx[:,1] -= self.gather("f_uy",type=1,count=1).flatten()
            dx[:,2] -= self.gather("f_uz",type=1,count=1).flatten()
            dx = self.pbc(dx)
            results.set("MaxDev",np.linalg.norm(dx,axis=1).max())
            results.set("Dev",dx.copy())
            del dx
            self.run_commands("unfix pafiax")
        
        # minimize to test for MaxJump
        if parameters("PostMin"):
            min_steps = parameters("MinSteps")
        else:
            min_steps = 1
        self.run_commands(f"""
            min_style fire
            minimize 0 0.0001 {min_steps} {min_steps}
        """)
        self.change_x += self.gather("x",1,3)
        results.set("MaxJump",self.pbc_dist(self.change_x,axis=1).max())
        results.set("Valid",bool(results("MaxJump")<parameters("MaxJumpThresh")))
        del self.change_x
        # unfix hyperplane
        self.run_commands("unfix pafi")
        self.run_script("PostRun",results)
        # rescale back.... not sure if this is required
        r = results("ReactionCoordinate")
        self.initialize_hyperplane(r,0.0)
        return results
  
            



        





        





from typing import Any,List

class ResultsHolder:
    """PAFI ResultsHolder object, 
    a dictionary which contains input and output values

    Methods
    -------
    __call__
    has_key
    """
    def __init__(self) -> None:
        self.data = {}
    def items(self)-> Any: # TODO what is the type here?
        """return items for iteration
        Returns
        -------
        dictionary.items()
        """
        return self.data.items()
    def __call__(self,key:str)->Any:
        """_summary_

        Parameters
        ----------
        key : str
            dictionary key

        Returns
        -------
        Any
           dictionary value
        
        Raises
        ------
        ValueError
            If key not found
        """
        if self.has_key(key):
            return self.data[key]
        else:
            raise ValueError(f"No key {key} in Results!")
        
    def set(self,key:str,value:Any)->None:
        """Set key,value pair in internal data

        Parameters
        ----------
        key : str
        value : Any
        """
        self.data[key] = value

    def has_key(self,key:str)->bool:
        """Check if key exists

        Parameters
        ----------
        key : str
           key query

        Returns
        -------
        bool
            True if exists
        """
        return bool(key in self.data.keys())
    
    def set_dict(self,dict:dict)->None:
        """Write or overwrite to results from dictionary
        writes key,value pairs via set()

        Parameters
        ----------
        dict : dict
        """
        for key,value in dict.items():
            self.set(key,value)
    def get_dict(self,keys:List[str],blanks:None|str=None)->dict:
        """Returns dictionary with given keys 

        Parameters
        ----------
        keys : List[str]
            list of keys
        blanks : None or str
            If not None, return str for missing key, default None

        Returns
        -------
        dict
            dictionary with values determined from internal data
        """
        res = {}
        for key in keys:
            if not blanks is None:
                res[key] = self.data[key] if self.has_key(key) else blanks
            else:
                res[key] = self.__call__(key)
        return res


import unittest,os
SELF_DIR = os.path.dirname(os.path.abspath(__file__))
DUMP_DIR = os.path.join(SELF_DIR,"./")
SYSTEM_DIR = os.path.join(SELF_DIR,"systems")

class PAFITestCase(unittest.TestCase):
	def test_import(self):
		# Add a name to the test case
		"""
		Test case for importing necessary modules
		"""
		import numpy # just to check import
		import scipy # just to check import
		import pandas # just to check import
		from mpi4py import MPI
		from pafi.workers.LAMMPSWorker import wrappedlammps
		self.assertTrue(MPI)
		self.assertTrue(wrappedlammps)
		comm = MPI.COMM_WORLD
		rank = comm.Get_rank()
		size = comm.Get_size()
		
		lmp = wrappedlammps(comm=comm,\
					 cmdargs=["-log","none","-screen","none"])
		lmp.close()
		
		self.assertTrue(rank >= 0)
		self.assertTrue(size > 0)
		print(""" 

		""")

	def test_load_parameters(self,nowrite=True):
		# Add a name to the test case
		"""
		Test case for loading parameters
		"""
		from pafi import PAFIParser
		self.assertTrue(PAFIParser)
		self.parameters = PAFIParser(postprocessing=nowrite)
		self.parameters.set("DumpFolder",DUMP_DIR)
		# set path wildcard for Fe SIA
		self.parameters.set_pathway(\
			os.path.join(SYSTEM_DIR,"EAM-SIA-Fe/image_*.dat"))
		# set EAM interatomic potential
		self.parameters.set_potential(\
			os.path.join(SYSTEM_DIR,"EAM-SIA-Fe/Fe.eam.fs"))
		# set potential species
		self.parameters.set_species(["Fe"])
		# restrict to zero temperature for testing 
		# TODO: finite temperature tests!
		self.parameters.axes["Temperature"] = [0.]
		# single force call for thermalization and sampling at T=0K
		self.parameters.set("nRepeats",1)
		self.parameters.set("Verbose",0)
		self.parameters.set("SampleSteps",1)
		self.parameters.set("ThermSteps",1)
		self.parameters.set("ThermWindow",1)
		self.assertTrue(self.parameters.ready())
		print(""" 




		""")
	
	def test_zero_temperature_run(self):
		# Add a name to the test case
		"""
		Test case for running at zero temperature
		"""
		import numpy as np
		from mpi4py import MPI
		from pafi import PAFIManager
		self.assertTrue(PAFIManager)
		self.test_load_parameters(False)
		manager = PAFIManager(world=MPI.COMM_WORLD,parameters=self.parameters)
		results = manager.run()
		manager.close()

		test_gradient = np.array([0.,0.614145,0.846192,0.59226,0.089912,
							 -0.426347,-0.853405, -0.792713, 0.])


		sim_gradient = np.array(results['FreeEnergyGradient'])
		
		self.assertTrue(np.isclose(sim_gradient,test_gradient).all())
		print(""" 




		""")

if __name__ == '__main__':
	unittest.main()



import numpy as np
import os,glob
import xml.etree.ElementTree as ET
import numpy as np
from typing import Union,Any,List
ScriptArg = Union[int,float,str]
from ..results.ResultsHolder import ResultsHolder

class BaseParser:
    """
    """
    def __init__(self,
                xml_path:None|os.PathLike[str]=None,
                postprocessing:bool=False,
                rank=0) -> None:
        """Base reader of PAFI XML configuration file
        
        Parameters
        ----------
        xml_path : os.PathLike[str]
            path to XML file, default None
        postprocessing: bool, optional
            are we just looking in postprocess?, default False
        
        Methods
        ----------
        __call__
        find_suffix_and_write
        

        Raises
        ------
        IOError
            If path is not found
        """
        self.postprocessing = postprocessing
        self.set_default_parameters()
        self.set_default_axes(empty=not xml_path is None)
        self.set_default_pathway()
        self.set_default_scripts()
        self.xml_path = xml_path
        self.rank=rank
        self.PotentialLocation = None
        self.Species = None
        if not xml_path is None:    
            assert os.path.exists(xml_path)
            xml_tree = ET.parse(xml_path)
            for branch in xml_tree.getroot():
                if branch.tag=="Axes":
                    self.read_axes(branch)
                elif branch.tag=="Parameters":
                    self.read_parameters(branch)
                elif branch.tag=="PathwayConfigurations":
                    self.read_pathway(branch)
                elif branch.tag=="Scripts":
                    self.read_scripts(branch)
                else:
                    raise IOError("Error in XML file!")
                if self.postprocessing:
                    self.suffix = int(xml_path.split("_")[-1].split(".")[0])
                    self.has_suffix=True
        else:
            if self.postprocessing:
                # for test routine
                print("Testing PAFI parser")
                self.suffix=""
                self.has_suffix=True
        self.suffix = -1
        self.xml_file = None
        
    def check(self)->None:
        """Internal check of parameters
        """
        if self.has_path and self.has_potential and self.has_species:
            self.check_output_location()
            self.check_axes()
            return True
        return False
    
    def ready(self)->bool:
        """Check if all parameters are set

        Returns
        -------
        bool
        """
        if not self.has_path:
            raise TypeError("\n\nNo Pathway set!\n")
        if not self.has_potential:
            raise TypeError("\n\nNo Potential set!\n")
        if not self.has_species:
            raise TypeError("\n\nNo Species set!\n")
        if not self.has_suffix:
            raise TypeError("\n\nNo output location set!\n")
        return True
        
    def check_axes(self)->None:
        """
            Ensure we have minimal axes for PAFI
        """
        assert "ReactionCoordinate" in self.axes.keys()
        assert "Temperature" in self.axes.keys()
    
    def check_output_location(self)->None:
        """
            Ensure DumpFolder exists and determine suffix
        """
        # dump data
        df = self.parameters["DumpFolder"]
        self.found_output_dir = os.path.isdir(df)
        if not self.found_output_dir:
            try:
                os.mkdir(df)
            except Exception as e:
                print(f"DumpFolder {df} cannot be made!",e)
        
        self.suffix=0
        self.find_suffix_and_write()
    
    def set_default_axes(self,empty=True) -> None:
        """Set defaults for "Axes" attribute

        Parameters
        ----------
        empty : bool, optional, by default True
            If True, we leave this blank to allow the
            configuration file to set axis order
        """
        self.axes = {}
        if not empty:
            self.axes["Temperature"] = np.zeros(1)
            self.axes["ReactionCoordinate"] = np.linspace(0.0,1.0,9)

        
        
    def read_axes(self,xml_axes:ET.Element) -> None:
        """Read "Axes" of XML file
            If data is in form `min max nsteps`, use `np.linspace`. 
            Otherwise, cast as `np.array`
        
        Parameters
        ----------
        xml_axes : xml.etree.ElementTreeElement
            Relevant branch of XML, as ElementTree Element
        """
        for axis in xml_axes:
            data = np.fromstring(axis.text.strip(),sep=' ')
            # check for form min max nsteps
            min_max_form = data.size==3 and data[0]<data[1]
            if min_max_form:
                min_max_form *= int(data[2])>1 and int(data[2])<20
            if min_max_form:
                self.axes[axis.tag] = np.linspace(data[0],data[1],int(data[2]))
            else:
                self.axes[axis.tag] = data.copy()

    def to_dict(self) -> dict:
        """Export axes and parameters as a nested dictionary

        Returns
        -------
        dict
            Dictionary-of-dictionaries with two keys 'axes' and 'parameters'
        """
        out_dict = {}
        out_dict['axes'] = self.axes.copy()
        out_dict['parameters'] = self.parameters.copy()
        return out_dict
    
    def set_default_parameters(self) -> None:
        """Set default values for <Parameters> data
        read_parameters() will *only* overwrite these values
        """
        self.parameters = {}
        self.parameters["CoresPerWorker"] = 1
        self.parameters["WriteDev"] = 0
        self.parameters["Verbosity"] = 0
        self.parameters["SampleSteps"] = 2000
        self.parameters["ThermSteps"] = 1000
        self.parameters["ThermWindow"] = 100
        self.parameters["MinSteps"] = 1000
        self.parameters["nRepeats"] = 1
        self.parameters["DumpFolder"] = './dumps'
        self.parameters["OverDamped"] = 0
        self.parameters["Friction"] = 0.05
        self.parameters["LogLammps"] = 0
        self.parameters["MaxJumpThresh"] = 0.4
        self.parameters["ReSampleThresh"] = 0.5
        self.parameters["maxExtraRepeats"] = 1
        self.parameters["PostDump"] = 0
        self.parameters["PreMin"] = 1
        self.parameters["PostMin"] = 1
        self.parameters["Verbose"] = 0
        self.parameters["SplinePath"] = 1
        self.parameters["RealMEPDist"] = 1
        self.parameters["GlobalSeed"] = 137
        self.parameters["FreshSeed"] = 1
        self.parameters["ReDiscretize"] = 1
        self.parameters["LinearThermalExpansion"] = np.zeros(3)
        self.parameters["QuadraticThermalExpansion"] = np.zeros(3)
        self.parameters["CubicSplineBoundaryConditions"] = "not-a-knot"
    
    def read_parameters(self,xml_parameters:ET.Element) -> None:
        """Read in simulation parameters defined in the XML file 

        Parameters
        ----------
        xml_parameters : xml.etree.ElementTreeElement
            The <Parameters> branch of the configuration file,
            represented as an ElementTree Element
        
        """
        for var in xml_parameters:
            tag = var.tag.strip()
            if not tag in self.parameters:
                print(f"Undefined parameter {tag}!!, skipping")
                continue
            else:
                o = self.parameters[tag]
                n = var.text
                if isinstance(o,int):
                    self.parameters[tag] = int(n)
                elif isinstance(o,float):
                    self.parameters[tag] = float(n)
                elif isinstance(o,bool):
                    self.parameters[tag] = int(n)
                elif isinstance(o,str):
                    self.parameters[tag] = n
                elif isinstance(o,list):
                    self.paramaters[tag] = n
                elif isinstance(o,np.ndarray):
                    nn = []
                    for _n in n.strip().split(" "):
                        if len(_n)>0:
                            nn += [_n]
                    nn = np.array(nn)
                    if "Expansion" in tag:
                        try:
                            nn = nn.astype(float)
                        except Exception as e:
                            print(f"Error in Expansion parameters",tag,e)
                        if nn.size == 1:
                            self.parameters[tag] = np.ones(3)*nn.sum()
                        elif nn.size==3:
                            self.parameters[tag] = nn.copy()
                        else:
                            print(f"Error in Expansion parameters")
                    else:
                        self.parameters[tag] = nn.copy()
        
    def __call__(self,key:str,value:Any=None) -> Any:
        """Extract or set the <Parameters> branch of the 
        PAFI configuration

        Parameters
        ----------
        key : str
            Field name of <Parameters>
        value : Any, optional
            if not None and key not found, set (key,value) pair, by default None

        Returns
        -------
        Any
            The value of the parameter
        """
        if not key in self.parameters:
            print(f"Key {key} not found, setting as {value}")
            self.parameters[key] = value
        return self.parameters[key]
    
    def has_key(self,key:str)->bool:
        """Check for parameter key

        Parameters
        ----------
        key : str
            key

        Returns
        -------
        bool
            result
        """
        return key in self.parameters.keys()
    
    def set_pathway(self,
                    paths:os.PathLike[str]|List[os.PathLike[str]],
                    directory:os.PathLike[str]="./") -> None:
        """Set the PathwayConfigurations
        
        Parameters
        ----------
        paths : os.PathLike[str]|List[os.PathLike[str]]
           wildcard or list of paths. 
           For wildcard, files are ordered according to INDEX
           assuming file format `any_form_of_path_INDEX.dat`

        directory : os.PathLike[str], optional
            starting path, by default "./"
        """
        assert isinstance(paths,list) or isinstance(paths,str)
        assert os.path.exists(directory)
        self.PathwayDirectory = directory
        
        if isinstance(paths,list) and len(paths)>1:
            paths = [os.path.join(directory.strip(),f.strip()) for f in paths]
        else:
            paths = paths[0] if isinstance(paths,list) else paths
            paths = glob.glob(os.path.join(directory.strip(),paths.strip()))
            paths = sorted(paths,key=lambda x:x.split("_")[-1])
        self.PathwayConfigurations = []
        for path in paths:
            try:
                assert os.path.exists(path)
            except AssertionError as e:
                print(path,e)
                raise AssertionError
            self.PathwayConfigurations += [path]
        self.has_path = True
        self.check()

    
    def set_potential(self,\
                    path:os.PathLike[str]|list[os.PathLike[str]],\
                    type="eam/fs",
                    species=None)->None:
        """Set potential pathway

        Parameters
        ----------
        path : os.PathLike[str] or list[os.PathLike[str]]
            path to potential file or list of files (for e.g. SNAP)
        """
        path = [path] if not isinstance(path,list) else path
        for p in path:
            if not os.path.exists(p):
                raise IOError(f"Potential {p} not found!")
        self.PotentialLocation = " ".join(path)
        self.PotentialType = type
        self.has_potential = True
        if not species is None:
            self.set_species(species)
        self.check()

    def set_species(self,species:str|List[str])->None:
        """Set element list
            Must be in correct order for LAMMPS to read !
            TODO: check this? 

        Parameters
        ----------
        species : str | List[str]
            string or list of species
        """
        if isinstance(species,str):
            species = [species]
        assert min([isinstance(s,str) for s in species])
        self.Species = species
        self.has_species = True
        self.check()

    def set_default_pathway(self) -> None:
        """
        Set default values for the <PathwayConfigurations> branch
        """
        self.has_path = False
        self.has_potential = False
        self.has_species = False
        self.has_suffix = False
        
    def read_pathway(self,xml_path_data:ET.ElementTree) -> None:
        """Read in pathway configuration paths defined in the XML file 

        Parameters
        ----------
        xml_path_data : xml.etree.ElementTreeElement
            The <PathwayConfigurations> branch of the configuration file,
            represented as an ElementTree Element
        """
        potential = None
        files = "*.dat"
        directory = "./"
        for path_data in xml_path_data:
            tag = path_data.tag.strip()
            if tag=="Potential":
                potential = path_data.text.strip()
            if tag=="Directory":
                directory = path_data.text.strip()
            if tag=="Files":
                files = path_data.text.strip().splitlines()
            if tag=="Species":
                species = path_data.text.strip()
        self.set_species(species)
        self.set_potential(potential)
        self.set_pathway(files,directory)
        
    
    def set_default_scripts(self) -> None:
        """Set default values for the <Scripts> branch
        """
        self.scripts={}
        self.scripts["Input"] = """
            units metal
            atom_style atomic
            atom_modify map array sort 0 0.0
            read_data  %FirstPathConfiguration%
            pair_style %PotentialType%
            pair_coeff * * %Potential% %Species%
            run 0
            thermo 10
            run 0
        """
        self.scripts["PreRun"] = """"""
        self.scripts["PostRun"] = """"""
        self.scripts["PreTherm"] = """"""
    
    def set_script(self,key:str,script:str) -> None:
        """Set a script

        Parameters
        ----------
        key : str
            key for <Script> in XML file
        script : str
            script
        """
        self.scripts[key] = script
    
    def read_scripts(self,xml_scripts) -> None:
        """Read in scripts defined in the XML file 

        Parameters
        ----------
        xml_parameters : xml.etree.ElementTreeElement
            The <Scripts> branch of the configuration file,
            represented as an ElementTree Element
        
        """
        for script in xml_scripts:
            if not script.text is None:
                tag = script.tag.strip()
                if not tag in self.scripts:
                    print(f"adding script {tag}")
                self.scripts[tag] = script.text.strip()
                
    
    def as_Element(self) -> ET.Element:
        """Cast all parameters as xml.etree.ElementTreeElement

        Returns
        -------
        xml.etree.ElementTree.Element
            Data as xml.etree.ElementTree
        """
        xmlET = ET.Element("PAFI")
        def add_branch(key:str,data:Any):
            """Add Axes, Parameter and Script data

            Parameters
            ----------
            key : str
                key
            data : Any
                data
            """
            branch = ET.Element(key)
            for key, val in data.items():
                child = ET.Element(key)
                if isinstance(val,list):
                    child.text = ""
                    for v in val:
                        child.text += str(v)+" "
                elif isinstance(val,np.ndarray):
                    child.text = ""
                    for v in val:
                        child.text += "%3.3g " % float(v)
                else:
                    child.text = str(val)
                branch.append(child)
            xmlET.append(branch)
        
        add_branch('Axes',self.axes)
        add_branch('Parameters',self.parameters)
        add_branch('Scripts',self.scripts)
        
        """
            Add PathwayConfiguration data
        """
        branch = ET.Element("PathwayConfigurations")
        branch_branch = ET.Element("Files")
        branch_branch.text = ""
        for p in self.PathwayConfigurations:
            branch_branch.text += os.path.basename(p) + "\n"
        branch.append(branch_branch)
        
        branch_branch = ET.Element("Potential")
        branch_branch.text = self.PotentialLocation
        branch.append(branch_branch)

        branch_branch = ET.Element("PotentialType")
        branch_branch.text = self.PotentialType
        branch.append(branch_branch)


        branch_branch = ET.Element("Species")
        branch_branch.text = " ".join(self.Species)
        branch.append(branch_branch)

        branch_branch = ET.Element("Directory")
        branch_branch.text = self.PathwayDirectory
        branch.append(branch_branch)
        xmlET.append(branch)

        
        
        return xmlET

    def to_string(self) -> str:
        """
            Return all paramaters as XML string
        """
        Element = self.as_Element()
        ET.indent(Element, '  ')
        return str(ET.tostring(Element)).replace('\\n','\n')
    
    def to_xml_file(self,xml_file:None|str=None)->None:
        """Write all paramaters as XML file
        
        Parameters
        ----------
        xml_file : str, optional
            path to XML file, default None
        """
        if xml_file is None:
            xml_file = self.xml_file
        assert not xml_file is None
        print(f"""
                Writing PAFI config file {xml_file}
                Will write PAFI data to {self.csv_file}
                """)
            
        Element = self.as_Element()
        ElementTree = ET.ElementTree(Element)
        ET.indent(ElementTree, '  ')
        ElementTree.write(xml_file,encoding="utf-8", xml_declaration=True)
        
    def replace(self,field:str,key:str,value: ScriptArg) -> str:
        """Wrapper around string replace()
           
        Parameters
        ----------
        field : str
            string to be searched
        key : str
            will search for %key%
        value : ScriptArg
            replacement value
        Returns
        -------
        str
            the string with replaced values
        """
        return field.replace("%"+key+"%",str(value))
    
    def parse_script(self,script_key:str,
                     arguments:None|dict|ResultsHolder=None) -> str:
        """Parse an input script
            If script_key is not a key of self.scripts, it 
            it is treated as a script itself

        Parameters
        ----------
        script_key : str
            key for <Script> in XML file
        args : None | dict, optional
            Dictionary of key,value pairs for replace(), by default None

        Returns
        -------
        str
            The script with any keywords replaced
        """
        if not script_key in self.scripts:
            script = script_key
        else:
            script = self.scripts[script_key]
        if arguments is None:
            _args = {}
        elif isinstance(arguments,ResultsHolder):
            _args = arguments.data.copy()
        else:
            _args = arguments.copy()
        _args["FirstPathConfiguration"] = self.PathwayConfigurations[0]
        if not self.PotentialLocation is None:
            _args["Potential"] = self.PotentialLocation
        if not self.PotentialType is None:
            _args["PotentialType"] = self.PotentialType
        if not self.Species is None:
            _args["Species"] = " ".join(self.Species)
        for key,value in _args.items():
            script = self.replace(script,key,value)
        return script
    
    
    def welcome_message(self):
        """
            A friendly welcome message :)
        """
        welcome = """
         _______      _______      _______       _________
        (  ____ )    (  ___  )    (  ____ \\    \\__   __/
        | (    )|    | (   ) |    | (    \\/       ) (
        | (____)|    | (___) |    | (__           | |
        |  _____)    |  ___  |    |  __)          | |
        | (          | (   ) |    | (             | |
        | )          | )   ( |    | )          ___) (___
        |/           |/     \\|    |/           \\_______/
        Projected    Average      Force        Integrator
            (c) TD Swinburne and M-C Marinica 2023
        
        Axes:
        """
        for key,val in self.axes.items():
            welcome += f"""
                {key}: {val}
            """
        
        welcome += """
        Pathway:
        """
        for path in self.PathwayConfigurations:
            welcome += f""" {path}
            """
        welcome += """
        Scripts:
        """
        for key,val in self.scripts.items():
            welcome += f"""
                {key}: 
                    {val}"""
        
        welcome += """
        Parameters:
        """
        for key,val in self.parameters.items():
            welcome += f"""
                {key}: {val}"""
        return welcome
    
    def find_suffix_and_write(self)->None:
        """
            Search output directory to find unique suffix
            for XML file log and CSV output data
        """

        if not self.has_suffix:
            if self.rank==0:
                df = self.parameters["DumpFolder"]
                fl = glob.glob(os.path.join(df,"config_*.xml"))
                self.suffix=-1
                for f in fl:
                    suffix = int(f.split("_")[-1][:-4])
                    self.suffix = max(self.suffix,suffix)
                self.suffix += 1
                self.xml_file = os.path.join(df,f"config_{self.suffix}.xml")
                self.csv_file = os.path.join(df,f"pafi_data_{self.suffix}.csv")
                self.to_xml_file()
            else:
                self.suffix = None
            self.has_suffix=True
                
                





from mpi4py import MPI
import os
import numpy as np
from typing import List
from ..parsers.PAFIParser import PAFIParser
from .ResultsHolder import ResultsHolder

class BaseGatherer:
    def __init__(self,params:PAFIParser,
                 nWorkers:int,
                 rank:int,
                 ensemble_comm:MPI.Intracomm,
                 roots:List[int])->None:
        """Basic gatherer of PAFI simulation data

        Parameters
        ----------
        params : PAFIParser
            Custom or predefined PAFIParser object
        nWorkers : int
            total number of PAFI workers
        rank : int
            global MPI rank
        ensemble_comm : MPI.Intracomm
            MPI communicator to gather ensemble data
        roots : List[int]
            list of root ranks for each worker. Only collate data here
            This should be depreceated asap
        """
        self.params = params
        self.nWorkers = nWorkers
        self.rank = rank
        self.comm = ensemble_comm
        self.roots = roots
        self.epoch_data = None # for each cycle
        self.all_data = None # for total simulation
        self.last_data = None # for print out
    
    def gather(self,data:dict|ResultsHolder)->None:
        """Gather results from a simulation epoch,
        local to each worker. Here, very simple,
        is overwritten by each call of gather()

        Parameters
        ----------
        data : dict or ResultsHolder
            Simulation data, extracted as dictionary from ResultsHolder
        """
        if self.rank in self.roots:
            if isinstance(data,ResultsHolder):
                self.epoch_data = data.data.copy()
            else:
                self.epoch_data = data.copy()
       
    
    def collate(self,repeat:int=0)->None:
        """Collate all data on root node
        It is assumed that all data is in dictionary form, with identical keys
        Only basic checks for multiple calls.

        Parameters
        ----------
        repeat : int, optional
            If  0, wipe last_data. Default 0
        """
        if self.rank in self.roots:
            if not self.epoch_data is None:
                all_epoch_data = self.comm.gather(self.epoch_data)
                self.epoch_data = None
            else:
                all_epoch_data = None

            if self.rank == 0 and not all_epoch_data is None:
                if (not repeat) or self.last_data is None:
                    self.last_data = {k:[] for k in all_epoch_data[0].keys()}
                if repeat:
                    self.last_data["Repeat"] = []

                if self.all_data is None:
                    self.all_data = {k:[] for k in all_epoch_data[0].keys()}
                
                for k in all_epoch_data[0].keys():
                    self.all_data[k] += list(d[k] for d in all_epoch_data)
                    self.last_data[k] += list(d[k] for d in all_epoch_data)

    
    def get_line(self,fields:List[str])->List[str]|None:
        """Return output data to print out on root node

        Parameters
        ----------
        fields : List[str]
            fields to extract
        Returns
        -------
        List[str]|None
            if root process, return list of lines to print, else return `None`
        """
        if self.rank != 0:
            return None
        line = []
        for f in fields:
            std = bool(f[-4:] == "_std")
            key = f[:-4] if std else f
            if (not self.last_data is None) and (key in self.last_data.keys()):
                d = self.last_data[key]
                line += [np.std(d)/np.sqrt(len(d)) if std else np.mean(d)]
            else:
                line += ["n/a"]
            
        return line
    
    def get_dict(self,fields:List[str])->dict|None:
        """Return output data to print out on root node

        Parameters
        ----------
        fields : List[str]
            fields to extract
        Returns
        -------
        dict|None
            if root process, return dict of fields to print, else return `None`
        """
        line = self.get_line(fields)
        if not line is None:
            return {kv[0]:kv[1] for kv in zip(fields,line)}
    
    def write_pandas(self,path:os.PathLike[str])->None:
        """Write data as pandas dataframe

        Parameters
        ----------
        path : os.PathLike[str]
            path to file
        """

        if self.rank==0:
            if self.all_data is None:
                print("No data to write! Exiting!")
            else:
                import pandas as pd
                if not os.path.isdir(os.path.dirname(path)):
                    raise IOError("Unknown directory for writing csv!")
                else:
                    df = pd.DataFrame(self.all_data)
                    df.metadata = " ".join(list(self.params.axes.keys()))
                    #df.attrs = self.params.to_dict().copy()
                    df.to_csv(path)
    
    def read_pandas(self,path:os.PathLike[str])->None:
        """Read in data from pandas dataframe

        Only possible if no sampling has taken place.

        Parameters
        ----------
        path : os.PathLike[str]
            path to file
        """
        assert self.all_data is None
        assert os.path.exists(path)
        import pandas as pd
        self.all_data = pd.read_csv(path)
            
                



import itertools
from typing import List,Dict
import numpy as np
import os
from mpi4py import MPI
from ..results.ResultsHolder import ResultsHolder
from .BaseManager import BaseManager
from ..parsers.PAFIParser import PAFIParser
from ..workers.PAFIWorker import PAFIWorker
from ..results.Gatherer import Gatherer

class PAFIManager(BaseManager):
    def __init__(self, world: MPI.Intracomm, 
                 xml_path:None|os.PathLike[str]=None,
                 parameters:None|PAFIParser=None,
                 restart_data:None|os.PathLike[str]=None,
                 Worker:PAFIWorker=PAFIWorker,
                 Gatherer:Gatherer=Gatherer) -> None:
        """Default manager of PAFI, child of BaseManager

        Parameters
        ----------
        world : MPI.Intracomm
            MPI communicator
        xml_path : None or os.PathLike[str], optional
            path to XML configuration file, default None
        parameters : None or PAFIParser object, optional
            preloaded PAFIParser object, default None
        restart_data : None or os.PathLike[str], optional
            path to CSV data file. Will read and skip already sampled parameters
        Worker : PAFIWorker, optional,
            Can be overwritten by child class, by default PAFIWorker
        Gatherer : Gatherer, optional
            Can be overwritten by child class, by default Gatherer
        """
        
        
        assert (not parameters is None) or (not xml_path is None)
        
        if parameters is None:
            # TODO have standalone check for suffix in config_[suffix].xml?
            # not the best solution currently if we use BaseManager alone....
            parameters = PAFIParser(xml_path=xml_path,rank=world.Get_rank())
        
        super().__init__(world, parameters, Worker, Gatherer)
        
    
    
    def run(self,print_fields:List[str]|None=None,
            width:int=10,
            precision:int=5)->None:
        """Basic parallel PAFI sampling

            Performs a nested loop over all <Axes>, in the order
            presented in the XML configuration file.
            Here, parallelization is naive- all workers are given the 
            same parameters.
        Parameters
        ----------
        print_fields : List[str] or None
            Fields to print to screen, default None. 
            If None, will print "Temperature","ReactionCoordinate","FreeEnergyGradient"
        width : int
            character count of field printout, default 10
        precision : int
            precision of field printout, default 4
        """
        assert self.parameters.ready()

        if print_fields is None:
            print_fields = \
                ["Temperature","postTemperature","ReactionCoordinate","FreeEnergyGradient","FreeEnergyGradient_std"]
        
        nRepeats = 1
        if not self.parameters("nRepeats") is None:
            if self.parameters("nRepeats")>1:
                nRepeats = self.parameters("nRepeats")
                print_fields = ["Repeat"] + print_fields
        
        for f in print_fields:
            width = max(width,len(f))
        
        def line(data:List[float|int|str]|Dict[str,float|int|str])->str:
            """Format list of results to print to screen
            """
            if len(data) == 0:
                return ""
            format_string = ("{: >%d} "%width)*len(data)
            if isinstance(data,dict):
                _fields = []
                for f in print_fields:
                    if f=='Repeat':
                        val = f"{int(data[f])}/{nRepeats}"
                    else:
                        val = data[f]
                    _fields += [val]
            else: 
                _fields = data
            
            fields = []
            for f in _fields:
                isstr = isinstance(f,str)
                fields += [f if isstr else np.round(f,precision)]
            return format_string.format(*fields)

        
        if self.rank==0:
            screen_out = f"""
            Initialized {self.nWorkers} workers with {self.CoresPerWorker} cores
            <> == time averages,  av/err over ensemble
            """
            if min(self.parameters.axes["Temperature"]) < 0.1:
                screen_out+="""
            *** FOR T=0K RUNS SampleSteps=1 AND ThermalSteps=1 ***
            """
            print(screen_out)
            print(line(print_fields))
            
            # return value
            average_results = {k:[] for k in print_fields}
            


        
        last_coord = None
        for axes_coord in itertools.product(*self.parameters.axes.values()):
            dict_axes = dict(zip(self.parameters.axes.keys(), axes_coord))
            
            if self.rank==0:
                if not last_coord is None and last_coord!=axes_coord[:-1]:
                    print("\n"+line(print_fields))
            if nRepeats>1:
                dict_axes["Repeat"] = 1
            results = ResultsHolder()
            results.set_dict(dict_axes)
            
            
            # Useful helper for including zero temperature cheaply...
            for k in ["SampleSteps","ThermSteps","ThermWindow"]:
                if results("Temperature")<0.1:
                    results.set(k,1)
                else:
                    results.set(k,self.parameters(k))
            
            for repeat in range(nRepeats):
                # Sampling run, returning ResultsHolder object
                final_results = self.Worker.sample(results)
                final_results.set("Repeat",repeat + 1)

                # incorporate results (this is only performed on local roots)
                self.Gatherer.gather(final_results)
                self.Gatherer.collate(repeat)

                # wait
                self.world.Barrier()
                screen_out = self.Gatherer.get_dict(print_fields)
                if self.rank == 0:
                    for k in screen_out.keys():
                        average_results[k].append(screen_out[k])
                    
                    print(line(screen_out))
                    self.Gatherer.write_pandas(path=self.parameters.csv_file)
            
            last_coord = axes_coord[:-1]
        
        if self.rank==0:
            print(f"Data written to {self.parameters.csv_file}")
            return average_results
        else:
            return None

        
    


import numpy as np
import os
from typing import List
from mpi4py import MPI
from ..parsers.PAFIParser import PAFIParser
from scipy.interpolate import CubicSpline

class BaseWorker:
    """Basic PAFI Worker

        Parameters
        ----------
        comm : MPI.Intracomm
            MPI communicator
        parameters : PAFIParser
            Predefined or custom  PAFIParser object
        worker_instance : int
            unique worker rank
        rank : int
            global rank (for MPI safety)
        roots : List[int]
            list of master ranks  (for MPI safety)
        """
    def __init__(self, comm : MPI.Intracomm,
                 parameters:PAFIParser,
                 worker_instance:int,
                 rank:int,
                 roots:List[int]) -> None:
        self.worker_instance = worker_instance
        self.comm = comm
        self.local_rank = comm.Get_rank()
        self.roots = roots
        self.rank = rank
        self.parameters = parameters
        self.error_count = 0
        self.scale = np.ones(3)
        self.out_width=16
        self.natoms=0
        self.nlocal=0
        self.offset=0
        self.x=None
        self.name="BaseWorker"
        self.has_errors = False
        self.has_cell_data=False
        self.Cell = None
        self.Periodicity = None
        self.invCell = None
        self.parameters.seed(worker_instance)
    
    def load_config(self,file_path:os.PathLike[str]) -> np.ndarray:
        """Placeholder function to load in file and return configuration
        Overwritten by LAMMPSWorker
        
        Parameters
        ----------
        file_path : os.PathLike[str]
            path to file

        Returns
        -------
        np.ndarray, shape (natoms,3)
            configuration vector
        """
        assert os.path.exists(file_path)
        return np.loadtxt(file_path)

    def pbc(self,X:np.ndarray,central:bool=True)->np.ndarray:
        """Minimum image convention, using cell data
            
        Parameters
        ----------
        X : np.ndarray, shape (natoms,3)
            configuration vector
        central : bool, optional
            map scaled coordinates to [-.5,.5] if True, else [0,1], by default True

        Returns
        -------
        np.ndarray
            wrapped vector
        """
        if not self.has_cell_data:
            return X
        else:
            sX = X.reshape((-1,3))@self.invCell
            sX -= np.floor(sX+0.5*int(central))@np.diag(self.Periodicity)
            return (sX@self.Cell).reshape((X.shape))
        
        return X
    def pbc_dist(self,X:np.ndarray,axis:None|int=None)->float|np.ndarray:
        """Minimum image distance
        
        Parameters
        ----------
        X : np.ndarray, shape (natoms,3)
            configuration vector
        axis : None | int, optional
            as in np.linalg.norm, by default None
        
        Returns
        -------
        float|np.ndarray
            norm of the vector or vector(s)
        """
        return np.linalg.norm(self.pbc(X),axis=axis)
    
    def make_path(self):
        """Make the splined PAFI path via scipy.interpolate.CubicSpline

            All parameters are read in from XML file

            Lots to do here in principle, TODO: parallel i/o and splining
            Only really a problem with memory limitations, say 2GB / core.
            This implies 300M double coordinates == 1M atoms, 100 planes.
            Typical large-scale use - 150k atoms, 20 planes.
            So memory-heavy but nothing problematic so far, leaving for future
        """

        # load configurations
        pc = self.parameters.PathwayConfigurations
        all_X = [self.pbc(self.load_config(pc[0]),central=False)]
        for p in pc[1:]:
            all_X += [self.pbc(self.load_config(p)-all_X[0])+all_X[0]]
            
        # determine distance TODO: symmetric??
        if self.parameters("RealMEPDist"):
            self.r_dist = np.array([self.pbc_dist(X-all_X[0]) for X in all_X])
            self.r_dist /= self.r_dist[-1]
        else:
            self.r_dist = np.linspace(0.,1.,all_X.shape[0])

        # splining - thank you scipy for 'axis' !!
        all_X = np.array([X.flatten() for X in all_X]) # shape=(nknots,natoms)
        bc = self.parameters("CubicSplineBoundaryConditions")
        assert bc in ['clamped','not-a-knot','natural']

        self.Spline_X = CubicSpline(self.r_dist,all_X,axis=0,bc_type=bc)
        del all_X # save a bit of memory

    def pathway(self,r:float,nu:int=0,
                scale:float|np.ndarray[3]=1.0)->np.ndarray:
        """Evaluate the PAFI pathway

        Parameters
        ----------
        r : float
            reaction coordinate, should be in [0,1]
        nu : int, optional
            derivative order, by default 0
        scale : float | np.ndarray[3], optional
            thermal expansion, by default 1.0

        Returns
        -------
        np.ndarray, shape (natoms,3)
            pathway configuration
        """
        if isinstance(scale,float):
            scale = np.eye(3)*float(scale)
        else:
            scale = np.diag(scale)
        return self.Spline_X(r,nu=nu).reshape((-1,3))@scale
    
    def close(self)->None:
        pass
            


        




import subprocess
import sys

def check_lammps():
    try:
        # Try to import lammps
        import lammps
        print("LAMMPS is installed and accessible.")
        return True
    except ImportError:
        print("LAMMPS is not installed or not accessible.")
        return False

def suggest_lammps_installation():
    print("\nTo install LAMMPS, please use conda:")
    print("conda install -c conda-forge lammps")
    print("\nAfter installing LAMMPS, you can install pafi using:")
    print("pip install pafi")

def main():
    if not check_lammps():
        suggest_lammps_installation()
        sys.exit(1)
    
    print("All dependencies are satisfied. You can proceed with using pafi.")

if __name__ == "__main__":
    main()


import numpy as np
import os
from typing import Any, List
from ..parsers.PAFIParser import PAFIParser
from mpi4py import MPI
from lammps import lammps,LMP_STYLE_GLOBAL,LMP_TYPE_VECTOR,LMP_TYPE_SCALAR
from .BaseWorker import BaseWorker
from ..results.ResultsHolder import ResultsHolder

class wrappedlammps(lammps):
    """
        Wrapper for lammps for safe MPI initialization

        Try to pass MPI communicator to lammps, if fails 
        we assume lammps is compiled without MPI support


    """
    def __init__(self,name='',cmdargs=None,ptr=None,comm=None):
        try:
            super().__init__(name=name,cmdargs=cmdargs,comm=comm)
            self.has_mpi_comm = True
        except Exception as ae:
            # in case we have conda-lammps installed 
            super().__init__(name=name,cmdargs=cmdargs,comm=None)
            self.has_mpi_comm = False
            # check if we are only assigning one core if no MPI support
            assert not self.lib.lammps_config_has_mpi_support()
            if comm is not None:
                assert comm.Get_size()==1



class LAMMPSWorker(BaseWorker):
    """LAMMPS worker for PAFI, inheriting BaseWorker

        Initialization routine:
        - runs "Input" script, loading first configuration
        - extracts cell data
        - defines various functions wrapping computes etc.
        - defines initialize_hyperplane function to set plane

        Parameters
        ----------
        comm : MPI.Intracomm
            MPI communicator
        parameters : PAFIParser
            Predefined or custom PAFIParser object
        worker_instance : int
            unique worker rank
        """
    def __init__(self, comm: MPI.Intracomm, 
                 parameters: PAFIParser, worker_instance: int,
                 rank: int, roots: List[int]) -> None:
        super().__init__(comm, parameters, worker_instance, rank, roots)
        
        self.name = "LAMMPSWorker"
        self.last_error_message = ""
        self.start_lammps()

        if (not self.L.has_mpi_comm) and comm.Get_size()>1:
            self.has_errors = True
            print("LAMMPS HAS NO MPI SUPPORT- ONE CORE/WORKER ONLY!")
            return
        
        if self.has_errors:
            print("ERROR STARTING LAMMPS!",self.last_error_message)
            return
        start_config = self.parameters.PathwayConfigurations[0]
        # TODO abstract
        self.run_script("Input")
        if self.has_errors:
            print("ERROR RUNNING INPUT SCRIPT!",self.last_error_message)
            return
        
        self.has_cell_data = False
        self.get_cell_data()
        # See initialize_hyperplane
        self.scale = np.ones(3)
        self.made_fix=False
        self.made_compute=False
        self.make_path()
        
    
    def start_lammps(self)->None:
        """Initialize LAMMPS instance

            Optionally 

        """
        if self.parameters("LogLammps"):
            logfile = 'log.lammps.%d' % self.worker_instance 
        else:
            logfile = 'none'
        try:
            cmdargs = ['-screen','none','-log',logfile]
            self.L = wrappedlammps(comm=self.comm,cmdargs=cmdargs)
            self.check_lammps_compatibility()
        except Exception as ae:
            print("Couldn't load LAMMPS!",ae)
            self.has_errors = True
    
    def check_lammps_compatibility(self)->None:
        """
            Ensure LAMMPS is new enough and has fix_pafi
        """
        lammps_release_int = self.L.version()
        if lammps_release_int<20201101: 
            if self.local_rank == 0:
                print("Require LAMMPS version > 01Nov2020!")
            return
        pafi_package = "EXTRA-FIX" if lammps_release_int>=20210728 else "USER-MISC"
        self.has_pafi = self.L.has_package(pafi_package)
        if not self.has_pafi and self.local_rank==0:
            print("Cannot find PAFI package in LAMMPS!")
            self.has_errors = True
        
    def run_script(self,key:str,arguments:None|dict|ResultsHolder=None)->None:
        """Run a script defined in the XML
            Important to replace any %wildcards% if they are there!
            
        Parameters
        ----------
        key : str
            script key from XML
        arguments : None | dict | ResultsHolder, optional
            will be used to replace wildcards, by default None
        """
        if key in self.parameters.scripts:
            if self.parameters("Verbose")>0 and self.rank==0:
                print(f"RUNNING SCRIPT {key}")
        
            script = self.parameters.parse_script(key,arguments=arguments)
            self.run_commands(script)

    def run_commands(self,cmds : str | List[str]) -> bool:
        """
            Run LAMMPS commands line by line, checking for errors
        """
        cmd_list = cmds.splitlines() if isinstance(cmds,str) else cmds
        for cmd in cmd_list:
            try:
                if self.parameters("Verbose")>0 and self.rank==0:
                    print(f"TRYING COMMAND {cmd}")
                self.L.command(cmd)
            except Exception as ae:
                if self.local_rank==0:
                    message = f"LAMMPS ERROR: {cmd} {ae}"
                else:
                    message = None
                self.last_error_message = ae
                raise SyntaxError(message)
    
    def gather(self,name:str,type:None|int=None,count:None|int=None)->np.ndarray:
        """Wrapper of LAMMPS gather()
           

        Parameters
        ----------
        name : str
            name of data
        type : None | int, optional
            type of array, 0:integer or 1:double, by default None.
            If None, an attempt will be made to determine autonomously.
        count : None | int, optional
            number of data per atom, by default None. 
            If None, an attempt will be made to determine autonomously.

        Returns
        -------
        np.ndarray
            the LAMMPS data

        Raises
        ------
        ValueError
            if name not found
        """
        if name in ['x','f','v'] or 'f_' in name:
            if type is None:
                type = 1
            if count is None:
                count = 3
        elif name in ['id','type','image']:
            if type is None:
                type = 0
            if count is None:
                count = 1
        if type is None or count is None:
            raise ValueError("Error in gather: type or count is None")
        
        try:
            res = self.L.gather(name,type,count)
        except Exception as ae:
            if self.local_rank==0:
                print("Error in gather:",ae)
            self.last_error_message = ae
        return np.ctypeslib.as_array(res).reshape((-1,count))
    
    def scatter(self,name:str,data:np.ndarray)->None:
        """Scatter data to LAMMPS
            Assume ordered with ID

        Parameters
        ----------
        name : str
            name of array
        data : np.ndarray
            numpy array of data. Will be flattened.
        """
        if np.issubdtype(data.dtype,int):
            type = 0
        elif np.issubdtype(data.dtype,float):
            type = 1
        count = data.shape[1] if len(data.shape)>1 else 1
        try:
            self.L.scatter(name,type,count,
                           np.ctypeslib.as_ctypes(data.flatten()))
        except Exception as ae:
            if self.local_rank==0:
                print("Error in scatter:",ae)
            self.last_error_message = ae
    
    def get_natoms(self)->int:
        """Get the atom count

        Returns
        -------
        int
            the atom count
        """
        return self.L.get_natoms()
    
    def get_cell_data(self)->None:
        """
            Extract supercell information
        """
        boxlo,boxhi,xy,yz,xz,pbc,box_change = self.L.extract_box()
        self.Periodicity = np.array([bool(pbc[i]) for i in range(3)],bool)
        self.Cell = np.zeros((3,3))
        for cell_j in range(3):
            self.Cell[cell_j][cell_j] = boxhi[cell_j]-boxlo[cell_j]
        self.Cell[0][1] = xy
        self.Cell[0][2] = xz
        self.Cell[1][2] = yz
        self.invCell = np.linalg.inv(self.Cell)
        self.has_cell_data = True
        return None

    def load_config(self,file_path:os.PathLike[str])->np.ndarray:
        """Load a LAMMPS data file with read_data() and return a numpy array of positions

        Parameters
        ----------
        file_path : os.PathLike[str]
            Path to the LAMMPS .dat file.

        Returns
        -------
        np.ndarray, shape (N,3)
            the positions
        """
        self.made_fix = False
        self.make_compute = False
        self.run_commands(f"""
            delete_atoms group all
            read_data {file_path} add merge
        """)
        return self.gather("x",type=1,count=3)

    def thermal_expansion_supercell(self,T:float=0) -> None:
        """Rescale the LAMMPS supercell *not coordinates*
            according to the provided thermal expansion data

        Parameters
        ----------
        T : float, optional
            temperature in K, by default 0
        """
        newscale = self.parameters.expansion(T)
        rs = newscale / self.scale
        self.run_commands(f"""
            change_box all x scale {rs[0]} y scale {rs[1]} z scale {rs[2]}
            run 0""")
        self.scale = newscale.copy()

    def extract_compute(self,id:str,vector:bool=True)->float|np.ndarray:
        """    Extract compute from LAMMPS
            
            id : str
        Parameters
        ----------
        id : str
            compute id
        vector : bool, optional
            is the return value a vector, by default True

        Returns
        -------
        np.ndarray or float
           return data
        """
        style = LMP_STYLE_GLOBAL
        type = LMP_TYPE_VECTOR if vector else LMP_TYPE_SCALAR
        assert hasattr(self.L,"numpy")
        try:
            res = self.L.numpy.extract_compute(id,style,type) 
            return np.array(res)
        except Exception as e:
            if self.local_rank==0:
                print("FAIL EXTRACT COMPUTE",e)
            self.close()
            return None
    
    def extract_fix(self,id:str,size:int=1)->float|np.ndarray:
        """Extract fix from LAMMPS and return a numpy array

        Parameters
        ----------
        id : str
            name of fix
        size : int, optional
            number of global entries, by default 1
        Returns
        -------
        float or np.ndarray
            numpy array of data of shape (size,), or float if size=1
        """
        
        style = LMP_STYLE_GLOBAL
        type = LMP_TYPE_VECTOR if size>1 else LMP_TYPE_SCALAR
        assert hasattr(self.L,"numpy")
        try:

            res = lambda i: self.L.numpy.extract_fix(id,style,type,ncol=i)
            if size>1:
                return np.array([res(i) for i in range(size)])
            else:
                return res(0)
        except Exception as e:
            if self.local_rank==0:
                print("FAIL EXTRACT FIX",id,e)
            self.close()
            return None
    
    def get_energy(self)->float:
        """Extract the potential energy
        
        Returns
        -------
        float
            The potential energy (pe)
        """
        return self.extract_compute("thermo_pe",vector=False)

    def initialize_hyperplane(self,r:float,T:float)->None:
        """Establish worker on a given plane with the 
            reference pathway via `fix_pafi`. 
            Also computes the tangent magnitude.

        Parameters
        ----------
        r : float
            The reaction coordinate, typically confined to [0,1]
        T : float
            The temperature in K. This will apply thermal expansion to the pathway.
        """
        
        self.thermal_expansion_supercell(T) # updates self.scale also

        # check for __pafipath fix to store the path data
        # (path : d_u[x,y,z],tangent: d_n[x,y,z],dtangent: d_dn[x,y,z])
        if not self.made_fix:
            self.run_commands(f"""
            fix __pafipath all property/atom d_ux d_uy d_uz d_nx d_ny d_nz d_dnx d_dny d_dnz
            run 0""")
            self.made_fix=True
        
        # fill positions: x and d_u[x,y,z]
        path_x = self.pathway(r,nu=0,scale=self.scale)
        self.scatter("x",path_x)
        for i,c in enumerate(["d_ux","d_uy","d_uz"]):
            self.scatter(c,path_x[:,i])
        del path_x

        # fill tangent: d_n[x,y,z]
        path_t = self.pathway(r,nu=1,scale=self.scale)
        path_t -= path_t.mean(0)
        self.norm_t = np.linalg.norm(path_t)
        path_t /= self.norm_t
        for i,c in enumerate(["d_nx","d_ny","d_nz"]):
            self.scatter(c,path_t[:,i])
        del path_t

        # fill dtangent: d_dn[x,y,z]
        path_t = self.pathway(r,nu=2,scale=self.scale) 
        
        path_t /= self.norm_t**2
        for i,c in enumerate(["d_dnx","d_dny","d_dnz"]):
            self.scatter(c,path_t[:,i])
        del path_t

        self.run_commands("run 0")

        # check for __pafipath compute to make this data accessible
        if not self.made_compute:
            self.run_commands(f"""
            compute __pafipath all property/atom d_ux d_uy d_uz d_nx d_ny d_nz d_dnx d_dny d_dnz
            run 0""")
            self.made_compute=True
    
    def close(self) -> None:
        """
            Close down. TODO Memory management??
        """
        super().close()
        self.L.close()

    


from .parsers.PAFIParser import PAFIParser
from .workers.PAFIWorker import PAFIWorker
from .results.ResultsHolder import ResultsHolder
from .results.ResultsProcessor import ResultsProcessor
from .managers.PAFIManager import PAFIManager


import numpy as np
from mpi4py import MPI
from typing import List
from ..parsers.PAFIParser import PAFIParser
from .BaseGatherer import BaseGatherer

class Gatherer(BaseGatherer):
    """Gatherer of PAFI simulation data

        Parameters
        ----------
        params : PAFIParser
            Custom or predefined PAFIParser object
        nWorkers : int
            total number of PAFI workers
        rank : int
            global MPI rank
        ensemble_comm : MPI.Intracomm
            MPI communicator to gather ensemble data
        roots : List[int]
            list of root ranks for each worker. Only collate data here
            This should be depreceated asap
    """
    def __init__(self, params: PAFIParser, nWorkers: int, rank: int, 
                 ensemble_comm: MPI.Intracomm, roots: List[int]) -> None:
        super().__init__(params, nWorkers, rank, ensemble_comm, roots)


from mpi4py import MPI
from ..parsers.PAFIParser import BaseParser
from ..workers.BaseWorker import BaseWorker
from ..results.BaseGatherer import BaseGatherer

class BaseManager:
    """Base class for PAFI manager

        Parameters
        ----------
        world : MPI.Intracomm
            MPI communicator
        parser : BaseParser object 
            A BaseParser or inherited class instance
        Worker : Worker class
            a predefined or custom Worker classes, default BaseWorker
        Gatherer : Gatherer class
            a predefined or custom Gatherer classes, default BaseGatherer
    """
    def __init__(self,world:MPI.Intracomm,parameters:BaseParser,
                 Worker=BaseWorker,Gatherer=BaseGatherer)->None:
        self.world = world
        self.rank = world.Get_rank()
        self.nProcs = world.Get_size()
        # Read in configuration file
        self.parameters = parameters
        self.CoresPerWorker = int(self.parameters("CoresPerWorker"))
        if self.nProcs%self.CoresPerWorker!=0:
            if self.rank==0:
                print(f"""
                    CoresPerWorker={self.CoresPerWorker} must factorize nProcs={self.nProcs}!!
                """)
                exit(-1)
        # Establish Workers
        # worker_comm : Worker communicator for e.g. LAMMPS
        self.nWorkers = self.nProcs // self.CoresPerWorker
        
        # Create worker communicator 
        self.worker_rank = self.rank // self.CoresPerWorker
        self.worker_comm = world.Split(self.worker_rank,0)
        
        # ensemble_comm: Global communicator for averaging
        self.roots = [i*self.CoresPerWorker for i in range(self.nWorkers)]
        self.ensemble_comm = world.Create(world.group.Incl(self.roots))
        

        # set up and seed each worker
        self.parameters.seed(self.worker_rank)
        self.Worker = Worker(self.worker_comm,
                             self.parameters,
                             self.worker_rank,
                             self.rank,
                             self.roots)
        
        
        # Establish Gatherer
        self.Gatherer = None
        if self.rank in self.roots:
            worker_errors = self.ensemble_comm.gather(self.Worker.has_errors)
            if self.rank==0 and max(worker_errors)>0:
                raise IOError("Worker Errors!")
            self.Gatherer = Gatherer(self.parameters,
                                 self.nWorkers,
                                 self.rank,
                                 self.ensemble_comm,
                                 self.roots)
        if self.rank==0:
            print(self.parameters.welcome_message())   
    
    def close(self)->None:
        """Close Manager
            closes Worker
        """
        self.Worker.close()
            


import pandas as pd
import os,itertools
import numpy as np
from typing import List
from ..parsers.PAFIParser import PAFIParser
from scipy.integrate import cumulative_trapezoid
from scipy.interpolate import interp1d

class ResultsProcessor:
    def __init__(self,
                 data_path:os.PathLike[str]|List[os.PathLike[str]],
                 xml_path:None|os.PathLike[str]=None,
                 axes:List[str]=None,
                 integrate:bool=True) -> None:
        """Read in PAFI data and plot results

        Parameters
        ----------
        data_path : os.PathLike[str] | List[os.PathLike[str]]
            path, wildcard, or list of paths to PAFI csv files
        xml_path : None | os.PathLike[str]
            path to PAFI XML configuration file, default None
        axes : List[str], optional
            List of axes, overwritten by data in xml_path if present,
            default : None, will be set to ["ReactionCoordinate","Temperature"]
        integrate : bool, optional
            perform integration with default arguments, default True
        Methods
        ----------
        append()
        extract_axes()
        integrate()
        """

        self.data = None
        self.axes = None
        self.fields = None
        if not isinstance(data_path,list):
            data_path = [data_path]
        for dp in data_path:
            if self.data is None:
                if os.path.exists(dp):
                    self.data = pd.read_csv(dp)
            else:
                self.append(dp)
        
        if (not xml_path is None) and os.path.exists(xml_path):
            self.params = PAFIParser(xml_path,postprocessing=True)
        else:
            self.params = None
        
        self.extract_axes(axes=axes)

        if self.integrate:
            self.integrate(return_remeshed_array=True,remesh=10);


    
    def extract_axes(self,axes=None):
        self.fields = list(self.data.keys())[1:]
        self.axes = {}

        if not self.params is None:
            axes = self.params.axes.keys()
        else:
            if axes is None:
                axes = ["ReactionCoordinate","Temperature"]
            print(f"""
                No config_*.xml file specified (via 'xml_path' argument)
                Using axes (sampling marginals): {axes}
                """)
        for a in axes:
            self.axes[a] = list(np.round(np.sort(np.unique(self.data[a])),4))

    def append(self,data_path)->None:
        if os.path.exists(data_path):
            new_data = pd.read_csv(data_path)
        
        """Append DataFrame to existing stored data. 
        Will throw error if `new_data` does not have fields of `self.data`stored fields to not cover append field

        Parameters
        ----------
        pd : pd.DataFrame
            DataFrame to append
        """
        if set(new_data.keys())>=set(self.data.keys()):
            old_data = self.data.copy()
            self.data = {}
            for f in old_data.keys():
                self.data[f] = \
                    pd.concat([old_data[f],new_data[f]],ignore_index=True)
            self.data = pd.DataFrame(self.data)
        else:
            print("Could not append!")
        
        # recheck ranges
        if not self.axes is None:
            for a in self.axes:
                self.axes[a] = \
                    list(np.round(np.sort(np.unique(self.data[a])),4))

    def ensemble_collate(self,return_pd:bool=False)->None|pd.DataFrame:
        """Collate ensemble data and calculate averages and variances.

        This method processes the stored data to calculate ensemble averages and variances
        for each unique combination of axes values. It handles valid/invalid data points
        and computes statistics only for valid entries.

        Parameters
        ----------
        return_pd : bool, optional
            If True, returns the collated data as a pandas DataFrame. Default is False.

        Returns
        -------
        pd.DataFrame or None
            If return_pd is True, returns a pandas DataFrame containing the collated data.
            Otherwise, returns None.

        Notes
        -----
        - The method uses the 'Valid' column to determine which data points to include in calculations.
        - Averages and variances are computed for all fields defined in self.fields.
        - The results are stored in self.ave_data as a pandas DataFrame.
        - A 'ValidCount' column is added to track the number of valid data points for each combination of axes values.
        """
        valid_key = 'Valid'
        self.count_key = 'ValidCount'
        axes = set(self.axes.keys())
        ave_data = {f:[] for f in self.fields}
        var_data = {f+"_var":[] for f in self.fields}
        count_data = {self.count_key:[]}
        
        # iterate over all axes
        for pt in itertools.product(*[self.axes[a] for a in axes]):
            sel = np.ones_like(self.data[valid_key].to_numpy(),bool)
            for a in zip(axes,pt):
                sel *= np.isclose(self.data[a[0]].to_numpy(),a[1])
            sel_data = self.data[sel]
            
            valid_r = sel_data[valid_key]
            valid_c = valid_r.sum()
            count_data[self.count_key] += [valid_c]

            for f in self.fields:
                if valid_c < 1:
                    ave_data[f] += [0.0]
                    var_data[f+"_var"] += [0.0]
                else:
                    valid_d = sel_data[f][valid_r].to_numpy()
                    ave_data[f] += [valid_d.mean()]
                    var_data[f+"_var"] += [valid_d.var()/valid_c]

        
        self.ave_data = pd.DataFrame({**ave_data,**count_data,**var_data})
        if return_pd:
            return self.ave_data

    def integrate(self,
                  argument:str='ReactionCoordinate',
                  target:str='FreeEnergyGradient',
                  variance:str='FreeEnergyGradientVariance',
                  remesh:int=5,
                  return_remeshed_array:bool=False)->pd.DataFrame:
        """Integrate the target field over the argument field.

        This method performs numerical integration of the target field (e.g., FreeEnergyGradient)
        with respect to the argument field (e.g., ReactionCoordinate). It also calculates
        the standard error of the integrated values.

        Parameters
        ----------
        argument : str, optional
            The field to use as the x-axis for integration (default is 'ReactionCoordinate')
        target : str, optional
            The field to integrate (default is 'FreeEnergyGradient')
        variance : str, optional
            The field containing the variance of the target (default is 'FreeEnergyGradientVariance')
        remesh : int, optional
            Factor by which to increase the density of integration points (default is 5)
        return_remeshed_array : bool, optional
            If True, return a remeshed array of the integrated data (default is False)

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the integrated values and their standard errors

        If return_remeshed_array is True:
        list of dict
            A list of dictionaries, each containing the remeshed data for a set of auxiliary parameters

        Notes
        -----
        The integration is performed using cubic spline interpolation. The method handles
        multiple sets of data distinguished by auxiliary parameters (e.g., different temperatures).
        The standard error of the integrated values is calculated using error propagation.

        The integrated values are added to the DataFrame with keys:
        - {target}_integrated: The integrated values
        - {target}_integrated_err: The standard error of the integrated values
        - {target}_integrated_u: Upper bound of the 95% confidence interval
        - {target}_integrated_l: Lower bound of the 95% confidence interval
        """
        
        # redo ensemble average
        self.ensemble_collate(return_pd=False)
        data = self.ave_data.copy()
        x_key = argument
        y_key = target
        y_key_std = target+"_err"
        y_key_var = target+"_var"
        v_key = variance
        
        i_key = target+"_integrated"
        i_key_std = target+"_integrated_err"
    
        i_keys = [i_key,i_key_std,i_key+"_u",i_key+"_l"]
        i_data = np.zeros((data[x_key].size,4))

        remesh = max(1,remesh)

        # sanity check!
        assert x_key in self.axes
        assert y_key in self.ave_data.keys()
        
        # we will integrate over constant aux values
        auxs = list(set(self.axes) - set([x_key]))
        
        if return_remeshed_array:
            out_array = []
        
        for pt in itertools.product(*[self.axes[a] for a in auxs]):
            # dataframe of selected data
            sel = np.ones_like(data[x_key].to_numpy(),bool)

            for a in zip(auxs,pt):
                sel *= np.isclose(data[a[0]].to_numpy(),a[1])
            sel_data = data[sel]
            
            # find integration points, sorted
            x_val = np.sort(np.unique(sel_data[x_key]))
            
            # for each x, find y and v values
            y_val = np.zeros((x_val.size,2))
            for i,x in enumerate(x_val):
                x_sel_data = sel_data[np.isclose(sel_data[x_key],x)]
                y_val[i][0] = float(x_sel_data[y_key].iloc[0])
                y_val[i][1] = float(x_sel_data[y_key_var].iloc[0])
                y_val[i][1] += float(x_sel_data[v_key].iloc[0])
                

            y_spl = interp1d(x_val,y_val,axis=0,kind='cubic')
            dense_x = np.linspace(x_val.min(),x_val.max(),remesh*x_val.size)
            dense_y = y_spl(dense_x)
            
            dense_i = cumulative_trapezoid(dense_y,dense_x,axis=0,initial=0)

            i_spl = interp1d(dense_x,dense_i,axis=0,kind='cubic')
            
            if return_remeshed_array:
                out_row = {a[0]:a[1] for a in zip(auxs,pt)}
                out_row[x_key] = dense_x
                out_row[y_key] = dense_y[:,0]
                out_row[y_key_var] = dense_y[:,1]
                out_row[y_key_std] = np.sqrt(np.abs(dense_y[:,1]))
                out_row[i_key] = dense_i[:,0]
                out_row[i_key_std] = np.sqrt(np.abs(dense_i[:,1]))
                out_array += [out_row]

            for i,x in enumerate(x_val):
                i_dat = i_spl(x)
                i_row = [i_dat[0],i_dat[1],i_dat[0]+i_dat[1],i_dat[0]-i_dat[1]]
                i_data[sel*np.isclose(data[x_key],x)] = i_row
            
        
        for i,k in enumerate(i_keys):
            data[k] = i_data[:,i]
            if k=='FreeEnergyGradient_integrated':
                data['FreeEnergyBarrier'] = i_data[:,i]
            if k=='FreeEnergyGradient_integrated_err':
                data['FreeEnergyBarrierErrors'] = i_data[:,i]    
        
        self.FreeEnergyData = []
        for T in np.unique(data['Temperature']):
            dd = data[data['Temperature']==T]
            res = {'Temperature':T}
            res['FreeEnergyGradient'] = \
                dd['FreeEnergyGradient']
            res['FreeEnergyBarrier'] = \
                dd['FreeEnergyGradient_integrated']
            res['ReactionCoordinate'] = dd['ReactionCoordinate']
            res['FreeEnergyBarrierError'] = \
                dd['FreeEnergyGradient_integrated_err']
            self.FreeEnergyData += [res]
        
        if return_remeshed_array:
            return data, out_array
        else:
            return data
    
    def plotting_data(self,remesh=10):
        """Generate data suitable for plotting.

        This method integrates the free energy gradient data to produce free energy profiles,
        which are more suitable for visualization. It uses a remeshing technique to ensure
        smooth curves.

        Parameters
        ----------
        remesh : int, optional
            The factor by which to increase the density of points along the reaction coordinate.
            Default is 10.

        Returns
        -------
        list of dict
            A list of dictionaries, each containing the data for a single temperature.
            Each dictionary has keys:
            - 'Temperature': The temperature of this dataset.
            - 'ReactionCoordinate': The x-values (reaction coordinate).
            - 'FreeEnergyGradient_integrated': The y-values (integrated free energy).
            - 'FreeEnergyGradient_integrated_std': The standard deviation of the integrated free energy.

        str
            The key for the x-axis data in the returned dictionaries ('ReactionCoordinate').

        str
            The key for the y-axis data in the returned dictionaries ('FreeEnergyGradient_integrated').

        Notes
        -----
        This method internally calls the `integrate` method with specific parameters to generate
        the plotting data. The integration is performed over the reaction coordinate using the
        free energy gradient and its variance.
        """
        _,ploting_data = self.integrate(remesh=10,return_remeshed_array=True,\
                argument='ReactionCoordinate',\
                target='FreeEnergyGradient',\
                variance='FreeEnergyGradientVariance')
        return ploting_data,'ReactionCoordinate','FreeEnergyGradient_integrated'


import numpy as np
import os
import numpy as np
from typing import Union,Any
ScriptArg = Union[int,float,str]

from .BaseParser import BaseParser

class PAFIParser(BaseParser):
    """Default reader of PAFI XML configuration file
    
    Parameters
    ----------
    xml_path : os.PathLike[str]
        path to XML file, default None
    postprocessing: bool, optional
        are we just looking in postprocess?, default False
    Methods
    ----------
    __call__
    find_suffix_and_write
    

    Raises
    ------
    IOError
        If path is not found
    """
    def __init__(self,
            xml_path:None|os.PathLike[str]=None,
            postprocessing:bool=False,
            rank:int=0) -> None:
        super().__init__(xml_path,postprocessing,rank)

        # initial seed, but must be different across workers...
        self.seeded = False
        
        # repeats, valid bounds (see set_min_valid())
        self.nRepeats = self.parameters["nRepeats"]
        self.maxExtraRepeats = self.parameters["maxExtraRepeats"]         
        self.set_min_valid(1) # temporary
        
    
    def set_min_valid(self,nWorkers:int)->None:
        """Set the minimum number of value samples 
        for each hyperplane

        Parameters
        ----------
        nWorkers : int
            number of MPI workers
        """
        
        self.minValidResults = self.parameters["ReSampleThresh"]
        self.minValidResults *= self.parameters["nRepeats"]
        self.minValidResults *= nWorkers
        self.minValidResults = int(self.minValidResults)
    def set(self,key:str,value:Any,create:bool=False)->None:
        """Set a parameter

        Parameters
        ----------
        key : str
            key of parameter. Must already exist if create is `False`
        value : Any
            value for entry
        create : bool, optional
            Create new entry if true    
        """
        if not create:
            assert key in self.parameters.keys()
        self.parameters[key] = value
    
    def min_valid(self)->int:
        """Return the minimum number of value samples, 
        set via set_min_valid()

        Returns
        -------
        int
            the return value
        """
        return self.minValidResults
    
    def seed(self,worker_instance:int)->None:
        """Generate random number seed

        Parameters
        ----------
        worker_instance : int
            unique to each worker, to ensure independent seed
        """
        if not self.seeded:
            self.randseed = self.parameters["GlobalSeed"] * (worker_instance+1)
            self.rng = np.random.default_rng(self.randseed)
            self.rng_int = self.rng.integers(low=100, high=10000)
            self.seeded=True
    
    def randint(self)->int:
        """Generate random integer.
            Gives exactly the same result each time unless reseed=True

        Returns
        -------
        int
            a random integer
        """
        if not self.seeded:
            print("NOT SEEDED!!")
            exit(-1)
        
        if self.parameters["FreshSeed"]:
            self.rng_int = self.rng.integers(low=100, high=10000)
        return str(self.rng_int)
    
    def expansion(self,T:float)->np.ndarray:
        """Return the relative x,y,z thermal expansion,
            using the data provided in the XML file

        Parameters
        ----------
        T : float
            Temperature in K

        Returns
        -------
        np.ndarray, shape (3,)
            relative x,y,z thermal expansion
        """
        scale = np.ones(3) 
        scale += self.parameters["LinearThermalExpansion"]*T
        scale += self.parameters["QuadraticThermalExpansion"]*T*T
        return scale
    
    def info(self)->str:
        """Return all parameters as formatted string

        Returns
        -------
        str
            the parameter string
        """
        result = """
            PAFI parameter information:
            
            Axes:"""
        for k,v in self.axes.items():
            result+=f"""
                {k} : 
                    {v}"""
        result += """
        
            Scripts:"""
        for k,v in self.scripts.items():
            result += f"""
                {k} : {self.parse_script(v)}"""
        result += f"""
            Wildcard Potential Type:
                {self.PotentialType}
            
            Wildcard Potential:
                {self.PotentialLocation}
            
            Wildcard Elements:
                {self.Species}

            Pathway:
                Directory:
                    {self.PathwayDirectory}
                Files:"""
        for p in self.PathwayConfigurations:
            result += f"""
                    {p}"""
        
        result += """


            Parameters"""
        for k,v in self.parameters.items():
            result += f"""
                {k} : {v}"""
        
        result += f"""
                seeded : {self.seeded}"""
        result += f"""
                minValidResults : {self.minValidResults}"""
        result += f"""

        """
        return result
    
    def to_dict(self) -> dict:
        """Export axes and parameters as a nested dictionary

        Returns
        -------
        dict
            Dictionary-of-dictionaries with two keys 'axes' and 'parameters'
        """
        return super().to_dict()
    
    

        

            


